<!DOCTYPE html>
<html>
<head>
    <title>RTV OAuth2 Helper</title>
    <!-- style borrowed from http://bettermotherfuckingwebsite.com/ -->
    <style type="text/css">
    body {
        margin:40px auto;
        max-width:650px;
        line-height:1.6;
        font-size:18px;
        font-family:Arial, Helvetica, sans-serif;
        color:#444;
        padding:0 10px;
    }
    h1, h2, h3 {
        line-height:1.2
    }
    #footer {
        position: absolute;
        bottom: 0px;
        width: 100%;
        font-size:14px;
    }
    </style>
</head>
<body>
${message}
    <div id="footer">View the <a href="http://www.github.com/michael-lazar/rtv">Documentation</a></div>
</body>
</html>
